import vision_models_playground.models
import vision_models_playground.components
import vision_models_playground.utility


__version__ = "0.1.0"
__author__ = "Alexandru Stirbu"
__email__ = "Alexandru.Stirbu.Net@outlook.com"
__license__ = "MIT"
__username__ = "Akrielz"
